{% skip_file_unless pyspark_job -%}

import bigflow

workflow_config = bigflow.Config(
    name='dev',
    is_master=True,
    properties={
        'gcp_project_id': {{ project_id|repr }},
        'staging_bucket': {{ project_name|repr }},
        'gcp_region': "europe-west1",
        'points': "100",
    },
)