#!/usr/bin/env python
import bigflow.build

bigflow.build.setup(
    name={{project_name | repr}},
    # author="unknown",
    # author_email="unknown@example.com",
    # description="bigflow project",
)
